use bevy::prelude::*;

use crate::ui::SelectedItem;
use crate::{CreateShape, ShapeData, ImageData};

#[derive(Debug, Clone)]
pub enum UndoItem {
    Base,
    CreateShape { entity: Entity },
    DeleteShape { sd: ShapeData, transform: Transform },
    MoveItem { shape: Entity, from: Vec2, to: Vec2 },
    MoveDot { shape_data: Entity, dot: Entity, from: Vec2, to: Vec2 },
    Rotate { entity: Entity, from: f32, to: f32 },
    NameChange { entity: Entity, from: String },
    NoteChange { entity: Entity, from: String },
    ChangeZ { entity: Entity, from: f32, to: f32 },
    ReScale { entity: Entity, from: Vec2, to: Vec2 },
    CreateImage { entity: Entity },
    DeleteImage { id: ImageData, transform: Transform, handle: Handle<Image> },
}
pub struct UndoStack {
    items: Vec<UndoItem>,
    current_action: usize,
    last_valid: usize,
}

impl Default for UndoStack {
    fn default() -> Self {
        Self { items: [UndoItem::Base].to_vec(), current_action: 0, last_valid: 0 }
    }
}
impl UndoStack {
    pub fn push(&mut self, item: UndoItem) {
        bevy::log::info!("PUSHING: {:?}", item);
        if self.current_action < self.items.len() - 1 {
            // We need to do a "semi push"
            self.current_action += 1;
            self.items[self.current_action] = item;
        }
        else {
            // We push a completly new item and might need more space!
            self.items.push(item);
            self.current_action = self.items.len() - 1;
        }
        self.last_valid = self.current_action;
    }
    /// Pop the last item in the stack
    pub fn pop(&mut self) -> Option<&mut UndoItem> {
        // If this happens we are clearly into invalid territory
        assert!(self.current_action < self.items.len());
        if self.current_action > 0 {
            let item = &mut self.items[self.current_action];
            self.current_action -= 1;
            Some(item)
        }
        else {
            None
        }
    }
    // If exists, "unpops" the last popped item
    pub fn unpop(&mut self) -> Option<&mut UndoItem> {
        if self.current_action < self.last_valid {
            self.current_action += 1;
            let item = &mut self.items[self.current_action];
            Some(item)
        }
        else {
            None
        }
    }
}

pub fn undo_redo_sys(
    mut coms: Commands,
    mut stack: ResMut<UndoStack>,
    mut inspector: ResMut<SelectedItem>,
    keyboard: Res<Input<KeyCode>>,
    mut shapes: Query<&mut ShapeData>,
    mut transforms: Query<&mut Transform>,
    mut images: Query<(&mut ImageData, &Handle<Image>)>,
) {
    if keyboard.just_pressed(KeyCode::Z) && keyboard.pressed(KeyCode::LControl) {
        let item = if keyboard.pressed(KeyCode::LShift) { stack.unpop() } else { stack.pop() };
        if let Some(item) = item {
            match item {
                UndoItem::Base => bevy::log::error!("POP/UNPOP: Got UndoItem::Base as a result!"),
                UndoItem::CreateShape { entity } => {
                    bevy::log::error!("NOT IMPLEMENTED : CreateShape");
                },
                UndoItem::DeleteShape {
                    sd,
                    transform,
                } => {
                    bevy::log::error!("NOT IMPLEMENTED : DeleteShape");
                },
                UndoItem::MoveItem { shape, from, to } => {
                    if let Ok(mut t) = transforms.get_mut(*shape) {
                        t.translation = from.extend(t.translation.z);
                    }
                    let t = *from;
                    *from = *to;
                    *to = t;
                },
                UndoItem::MoveDot { shape_data, dot, from, to } => {
                    bevy::log::error!("NOT IMPLEMENTED : MoveDot");
                },
                UndoItem::Rotate { entity, from, to } => {
                    if let Ok(mut t) = transforms.get_mut(*entity) {
                        t.rotation = Quat::from_rotation_z(from.to_radians());
                    }
                    let t = *from;
                    *from = *to;
                    *to = t;
                },
                UndoItem::NameChange { entity, from } => {
                    if let Ok(mut sd) = shapes.get_mut(*entity) {
                        std::mem::swap(&mut sd.name, from);
                    }
                    else if let Ok((mut id, _)) = images.get_mut(*entity) {
                        std::mem::swap(&mut id.name, from);
                    }
                    // No need to update the item, so no need to replace
                },
                UndoItem::ChangeZ { entity, from, to } => {
                    if let Ok(mut t) = transforms.get_mut(*entity) {
                        t.translation.z = *from;
                    }
                    let t = *from;
                    *from = *to;
                    *to = t;
                },
                UndoItem::ReScale { entity, from, to } => {
                    if let Ok(mut t) = transforms.get_mut(*entity) {
                        t.scale = from.extend(1.0);
                    }
                    let t = *from;
                    *from = *to;
                    *to = t;
                },
                UndoItem::CreateImage { entity } => {
                    if let Ok((mut id, handle)) = images.get_mut(*entity) {
                        let transform = transforms.get(*entity).map(|t| *t).unwrap_or(Transform::default());
                        let mut nid = ImageData::default();
                        std::mem::swap(&mut *id, &mut nid);

                        coms.entity(*entity).despawn();
                        *item = UndoItem::DeleteImage { id: nid, transform, handle: handle.clone() };
                    }
                },
                UndoItem::DeleteImage { transform, handle, id } => {
                    let mut nid = ImageData::default();
                    std::mem::swap(id, &mut nid);

                    let e = coms.spawn_bundle(SpriteBundle {
                        texture: handle.clone(),
                        transform: *transform,
                        ..Default::default()
                    })
                    .insert(nid).id();

                    *item = UndoItem::CreateImage { entity: e };
                },
                UndoItem::NoteChange { entity, from } => {
                    if let Ok(mut sd) = shapes.get_mut(*entity) {
                        std::mem::swap(&mut sd.note, from);
                    }
                    else if let Ok((mut id, _)) = images.get_mut(*entity) {
                        std::mem::swap(&mut id.note, from);
                    }
                },
            }
        }
        else {
            bevy::log::info!("Nothing to pop/unpop!");
        }
    }
}